#ifndef _OPENDRIVE_ROAD_SEGMENT_H
#define _OPENDRIVE_ROAD_SEGMENT_H

#ifdef _HAVE_XSD
#include "xml/OpenDRIVE_1.4H.hxx"
#endif

#include "opendrive_common.h"
#include "opendrive_lane.h"
#include "opendrive_signal.h"
#include "opendrive_object.h"
#include "opendrive_road.h"
#include "opendrive_road_node.h"
#include "opendrive_geometry.h"

typedef struct{
  COpendriveGeometry *opendrive;
  CG2Spline *spline;
}TOpendriveRoadSegmentGeometry;

class COpendriveRoadSegment
{
  friend class COpendriveRoad;
  friend class COpendriveRoadNode;
  friend class COpendriveLane;
  private:
    std::map<int,COpendriveLane *> lanes;
    std::vector<COpendriveRoadNode *> nodes;
    std::vector<TOpendriveRoadSegmentGeometry> geometries;
    COpendriveRoad *parent_road;
    double resolution;
    double scale_factor;
    double min_road_length;
    int num_right_lanes;
    int num_left_lanes;
    std::vector<COpendriveSignal *> signals;
    std::vector<COpendriveObject *> objects;
    std::vector<COpendriveRoadSegment *> connecting;
    std::string name;
    unsigned int id;
    opendrive_mark_t center_mark;
  protected:
    COpendriveRoadSegment();
    COpendriveRoadSegment(const COpendriveRoadSegment &object);
    void load(OpenDRIVE::road_type &road_info);
    void free(void);
    void set_resolution(double res);
    void set_scale_factor(double scale);
    void set_min_road_length(double length);
    void set_parent_road(COpendriveRoad *parent);
    void set_name(std::string &name);
    void set_id(unsigned int id);
    void set_center_mark(opendrive_mark_t mark);
    bool updated(segment_up_ref_t &segment_refs);
    COpendriveRoadSegment *get_original_segment(segment_up_ref_t &segment_refs);
    void update_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void clean_references(segment_up_ref_t &segment_refs,lane_up_ref_t &lane_refs,node_up_ref_t &node_refs);
    void add_lanes(lanes::laneSection_type &lane_section);
    void add_lane(const ::lane &lane_info);
    void remove_lane(COpendriveLane *lane);
    void add_geometries(OpenDRIVE::road_type &road_info);
    void add_nodes(void);
    void add_node(TOpendriveRoadSegmentGeometry &geometry,COpendriveLane *lane);
    void add_empty_node(TOpendriveWorldPose &pose,COpendriveLane *lane);
    void remove_node(COpendriveRoadNode *node);
    bool has_node(COpendriveRoadNode *node) const;
    void link_neightbour_lanes(lanes::laneSection_type &lane_section);
    void link_neightbour_lanes(void);
    void link_segment(COpendriveRoadSegment *segment);
    void link_segment(COpendriveRoadSegment *segment,int from,bool from_start, int to,bool to_start);
    bool connects_to(COpendriveRoadSegment *segment);
    bool connects_segments(COpendriveRoadSegment *segment1,COpendriveRoadSegment *segment2);
    COpendriveRoadSegment *get_sub_segment(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref,int node_side,TOpendriveWorldPose *start=NULL,TOpendriveWorldPose *end=NULL) const;
    COpendriveRoadSegment *clone(node_up_ref_t &new_node_ref,lane_up_ref_t &new_lane_ref,link_up_ref_t &new_link_ref) const;
  public:
    std::string get_name(void) const;
    unsigned int get_id(void) const;
    unsigned int get_num_right_lanes(void) const;
    unsigned int get_num_left_lanes(void) const;
    const COpendriveLane &get_lane(int index) const;
    unsigned int get_num_nodes(void) const;
    const COpendriveRoadNode &get_node(unsigned int index) const;
    const COpendriveRoad &get_parent_road(void) const;
    unsigned int get_num_signals(void) const;
    const COpendriveSignal &get_signal(unsigned int index) const;
    unsigned int get_num_objects(void) const;
    const COpendriveObject &get_object(unsigned int index) const;
    unsigned int get_num_connecting(void) const;
    const COpendriveRoadSegment &get_connecting(unsigned int index) const;
    double get_length(void) const;
    TOpendriveWorldPose get_pose_at(double length) const;
    double get_curvature_at(double length) const;
    TOpendriveLocalPose transform_to_local_pose(const TOpendriveTrackPose &track);
    TOpendriveWorldPose transform_to_world_pose(const TOpendriveTrackPose &track);
    const COpendriveRoadNode &get_closest_node(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    unsigned int get_closest_node_index(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    const COpendriveLane &get_closest_lane(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    int get_closest_lane_id(TOpendriveWorldPose &pose,double &distance,double angle_tol=0.1) const;
    double get_closest_pose(TOpendriveWorldPose &pose,TOpendriveWorldPose &closest_pose,double angle_tol=0.1) const;
    const COpendriveRoadSegment &get_next_segment(int &side,bool &valid) const;
    const COpendriveRoadSegment &get_prev_segment(int &side,bool &valid) const;
    int get_pose_side(TOpendriveWorldPose &pose) const;
    friend std::ostream& operator<<(std::ostream& out, COpendriveRoadSegment &segment);
    ~COpendriveRoadSegment();
};

#endif
