#include "opendrive_signal.h"
#include "exceptions.h"
#include <cmath>

COpendriveSignal::COpendriveSignal()
{
  this->segment=NULL;
  this->id=-1;
  this->pose.s=0.0;
  this->pose.t=0.0;
  this->pose.heading=0.0;
  this->type="";
  this->sub_type="";
  this->value=0;
  this->text="";
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

COpendriveSignal::COpendriveSignal(const COpendriveSignal &object)
{
  this->segment=object.segment;
  this->id=object.id;
  this->pose.s=object.pose.s;
  this->pose.t=object.pose.t;
  this->pose.heading=object.pose.heading;
  this->type=object.type;
  this->sub_type=object.sub_type;
  this->value=object.value;
  this->text=object.text;
  this->scale_factor=object.scale_factor;
}

void COpendriveSignal::load(signals::signal_type &signal_info,COpendriveRoadSegment *segment)
{
  std::stringstream ss(signal_info.id().get());
  ss >> this->id;
  this->pose.s = (signal_info.s().present() ? signal_info.s().get() : 0.0);
  this->pose.t = (signal_info.t().present() ? signal_info.t().get() : 0.0);
  this->pose.heading = (signal_info.hOffset().present() ? signal_info.hOffset().get() : 0.0);
  this->type = (signal_info.type().present() ? signal_info.type().get() : "");
  this->sub_type = (signal_info.subtype().present() ? signal_info.subtype().get() : "");
  this->value = (signal_info.value().present() ? signal_info.value().get() : 0.0);
  this->text = (signal_info.text().present() ? signal_info.text().get() : "");
  bool reverse = false;
  if (signal_info.orientation().present() && signal_info.orientation().get() == orientation::cxx_1)
    reverse = true;
  this->pose.heading = normalize_angle(this->pose.heading + (reverse ? M_PI : 0.0));
  this->segment=segment;
}

void COpendriveSignal::update_references(segment_up_ref_t &segment_refs)
{
  if(segment_refs.find(this->segment)!=segment_refs.end())
    this->segment=segment_refs[this->segment];
}

void COpendriveSignal::set_scale_factor(double scale)
{
  this->scale_factor=scale;
}

int COpendriveSignal::get_id(void) const
{
  return this->id;
}

double COpendriveSignal::get_scale_factor(void)
{ 
  return this->scale_factor;
}

TOpendriveTrackPose COpendriveSignal::get_track_pose(void) const
{
  TOpendriveTrackPose track;
 
  track.s=this->pose.s/this->scale_factor;
  track.t=this->pose.t/this->scale_factor;
  track.heading=this->pose.heading;

  return track;
}

TOpendriveWorldPose COpendriveSignal::get_world_pose(void) const
{
  TOpendriveWorldPose world;

  if(this->segment!=NULL)
  {
    world=segment->transform_to_world_pose(this->get_track_pose());
    return world;
  }
  else
    throw CException(_HERE_,"Invalid parent segment reference");
}

void COpendriveSignal::get_type(std::string &type, std::string &sub_type) const
{
  type=this->type;
  sub_type=this->sub_type;
}

int COpendriveSignal::get_value(void) const
{
  return this->value;
}

std::string COpendriveSignal::get_text(void) const
{
  return this->text;
}

std::ostream& operator<<(std::ostream& out, COpendriveSignal &signal)
{
  TOpendriveTrackPose track;

  out << "  Signal id = " << signal.id << std::endl;
  out << "    type = " << signal.type << "," << signal.sub_type << std::endl;
  out << "    value = " << signal.value << std::endl;
  out << "    text = " << signal.text << std::endl;
  track=signal.get_track_pose();
  out << "    pose: s = " << track.s << ", t = " << track.t << ", heading = " << track.heading << std::endl;

  return out;
}

COpendriveSignal::~COpendriveSignal()
{
  this->segment=NULL;
  this->id=-1;
  this->pose.s=0.0;
  this->pose.t=0.0;
  this->pose.heading=0.0;
  this->type="";
  this->sub_type="";
  this->value=0;
  this->text="";
  this->scale_factor=DEFAULT_SCALE_FACTOR;
}

